from typing import Optional
import pandas as pd
import numpy as np

class DataFrameBuilder:
    def __init__(self, columns):
        self._data = {c: [] for c in columns}

    def extend(self, values):
        assert all(c in values.keys() for c in self._data), f"all columns must be provided: {list(self._data)}"
        lens = {k: len(v) for k, v in values.items() if hasattr(v, '__len__') and not isinstance(v, str)}
        assert any(lens), "at least one value must have a length"
        first_len = list(lens.values())[0]
        assert all(l == first_len for l in lens.values()), f"all values must have the same length {lens}"
        for k, v in values.items():
            if k not in lens:
                self._data[k].append([v] * first_len)
            elif isinstance(v, pd.Series):
                self._data[k].append(v.values)
            else:
                self._data[k].append(v)

    def to_df(self, merge_on_index: Optional[pd.DataFrame] = None):
        result = pd.DataFrame({
            k: np.concatenate(v)
            for k, v in self._data.items()
        })
        if merge_on_index is not None:
            assert '_index' in result
            merge_on_index = merge_on_index.reset_index(drop=True)
            result = result.assign(**{
                col: merge_on_index[col].iloc[result['_index']].values
                for col in merge_on_index.columns
                if col not in result.columns
            }).drop(columns=['_index'])
            merge_columns = set(merge_on_index.columns)
            column_order = list(merge_on_index.columns) + [c for c in result.columns if c not in merge_columns]
            result = result[column_order]
        return result



from typing import Optional
import ir_measures


def RBO(other, p=0.99, *, name: Optional[str] = None):
    return ir_measures.define(_rbo_wrapper(other, p=0.999), name=name or f'RBO(p={p})')


def _rbo_wrapper(a, p=0.99):
    # adapted from https://github.com/terrierteam/ir_measures/blob/main/ir_measures/providers/compat_provider.py
    a_q_col = 'query_id' if 'query_id' in a.columns else 'qid'
    a_d_col = 'doc_id' if 'doc_id' in a.columns else 'docno'
    a = a.sort_values(by=[a_q_col, 'score'], ascending=False)
    a = dict(iter(a.groupby(a_q_col)))
    def inner(qrels, b):
        # qrels ignored
        b_q_col = 'query_id' if 'query_id' in b.columns else 'qid'
        b_d_col = 'doc_id' if 'doc_id' in b.columns else 'docno'
        res = {}
        b = b.sort_values(by=[b_q_col, 'score'], ascending=False)
        b = dict(iter(b.groupby(b_q_col)))
        for qid in set(a.keys()) | set(b.keys()):
            ranking = list(a[qid][a_d_col]) if qid in a else []
            ideal = list(b[qid][b_d_col]) if qid in b else []
            ranking_set = set()
            ideal_set = set()
            score = 0.0
            normalizer = 0.0
            weight = 1.0
            for i in range(1000):
                if i < len(ranking):
                    ranking_set.add(ranking[i])
                if i < len(ideal):
                    ideal_set.add(ideal[i])
                score += weight*len(ideal_set.intersection(ranking_set))/(i + 1)
                normalizer += weight
                weight *= p
            res[qid] = score/normalizer
        return res.items()
    return inner


def rbo(a, b, p=0.99):
    return _rbo_wrapper(a, p)(b)



import contextlib
import json
import os
import tarfile
from pathlib import Path
from hashlib import sha256
from typing import Dict, Optional, Union
from urllib.parse import urlparse
import tempfile

from lz4.frame import LZ4FrameFile

import pyterrier_alpha as pta


class Artifact:
    """Base class for PyTerrier artifacts.

    An artifact is a component stored on disk, such as an index.

    Artifacts usually act as factories for transformers that use them. For example, an index artifact
    may provide a `.retriever()` method that returns a transformer that searches the index.
    """
    def __init__(self, path: Union[Path, str]):
        self.path: Path = Path(path)

    @classmethod
    def load(cls, path: str) -> 'Artifact':
        """Load the artifact from the specified path.

        If invoked on the base class, this method will attempt to find a supporting Artifact
        implementation that can load the artifact at the specified path. If invoked on a subclass,
        it will attempt to load the artifact using the specific implementation.

        Args:
            path: The path of the artifact on disk.

        Returns:
            The loaded artifact.

        Raises:
            FileNotFoundError: If the specified path does not exist.
            ValueError: If no implementation is found that supports the artifact at the specified path.
        """
        if cls is Artifact:
            return load(path)
        else:
            # called SomeArtifact.load(path), so load this specific artifact
            # TODO: add error message if not loaded
            return cls(path)

    @classmethod
    def from_url(cls, url: str, *, expected_sha256: Optional[str] = None) -> 'Artifact':
        """Load the artifact from the specified URL.

        The artifact at the specified URL will be downloaded and stored in PyTerrier's artifact cache.

        Args:
            url: The URL or file path of the artifact.

        Returns:
            The loaded artifact.

        Raises:
            ValueError: If no implementation is found that supports the artifact at the specified path.
        """

        # Support mapping "protocols" of the URL other than http[s]
        # e.g., "hf:abc/xyz@branch" -> "https://huggingface.co/datasets/abc/xyz/resolve/branch/artifact.tar"
        parsed_url = urlparse(url)
        if parsed_url.scheme == 'hf':
            org_repo = parsed_url.path
            # default to ref=main, but allow user to specify another branch, hash, etc, with abc/xyz@branch
            ref = 'main'
            if '@' in org_repo:
                org_repo, ref = org_repo.split('@', 1)
            url = f'https://huggingface.co/datasets/{org_repo}/resolve/{ref}/artifact.tar.lz4'
            parsed_url = urlparse(url)

        # buid local path
        base_path = os.path.join(pta.io.pyterrier_home(), 'artifacts')
        if not os.path.exists(base_path):
            os.makedirs(base_path)
        path = os.path.join(base_path, sha256(url.encode()).hexdigest())

        if not os.path.exists(path):
            # try to load the package metadata file (if it exists)
            download_info = {}
            try:
                with pta.io.open_or_download_stream(f'{url}.json', verbose=False) as fin:
                    download_info = json.load(fin)
                if 'expected_sha256' in download_info:
                    new_expected_sha256 = download_info['expected_sha256'].lower()
                    if expected_sha256 is not None:
                        assert expected_sha256.lower() == new_expected_sha256
                    expected_sha256 = new_expected_sha256
                if 'total_size' in download_info:
                    # TODO: make sure there's available space on the disk for this
                    pass
            except OSError:
                pass


            with contextlib.ExitStack() as stack:
                dout = stack.enter_context(pta.io.finalized_directory(path))

                metadata_out = stack.enter_context(pta.io.finalized_open(f'{path}.json', 't'))
                if parsed_url.scheme == '':
                    json.dump({'path': url}, metadata_out)
                else:
                    json.dump({'url': url}, metadata_out)
                metadata_out.write('\n')

                if 'segments' in download_info:
                    # the file is segmented -- use a sequence reader to stitch them back together
                    fin = stack.enter_context(pta.io.BufferedSequenceReader(
                        pta.io.open_or_download_stream(f'{url}.{i}')
                        for i in range(len(download_info['segments']))
                    ))
                    if expected_sha256 is not None:
                        fin = stack.enter_context(pta.io.HashReader(fin, expected=expected_sha256))
                else:
                    hash_fin = None # hash verification handled by open_or_download_stream
                    fin = stack.enter_context(pta.io.open_or_download_stream(url, expected_sha256=expected_sha256))

                cin = fin
                if parsed_url.path.endswith('.lz4'):
                    cin = stack.enter_context(LZ4FrameFile(fin))
                # TODO: support other compressions

                tar_in = stack.enter_context(tarfile.open(fileobj=cin, mode='r|'))

                metadata_out.flush()
                for member in tar_in:
                    if (member.isfile() or member.isdir()) and pta.io.path_is_under_base(member.path, dout):
                        print(f'extracting {member.path} [{pta.io.byte_count_to_human_readable(member.size)}]')
                        tar_in.extract(member, dout, set_attrs=False)

        return cls.load(path)

    @classmethod
    def from_hf(cls, hf_repo: str, branch: str = 'main', *, expected_sha256: Optional[str] = None) -> 'Artifact':
        return cls.from_url(f'hf:{hf_repo}@{branch}', expected_sha256=expected_sha256)

    def to_hf(self, repo, branch='main'):
        import huggingface_hub
        with tempfile.TemporaryDirectory() as d:
            # build a package with a maximum individual file size of just under 5GB, the limit for HF datasets
            build_package(self.path, os.path.join(d, 'artifact.tar.lz4'), max_file_size=4.9e9)
            try:
                huggingface_hub.create_repo(repo, repo_type='dataset')
            except huggingface_hub.utils.HfHubHTTPError as e:
                if e.server_message != 'You already created this dataset repo':
                    raise
            try:
                huggingface_hub.create_branch(repo, repo_type='dataset', branch=branch)
            except huggingface_hub.utils.HfHubHTTPError as e:
                if not e.server_message.startswith('Reference already exists:'):
                    raise
            huggingface_hub.upload_folder(
                repo_id=repo,
                folder_path=d,
                repo_type='dataset',
                revision=branch,
            )

    @classmethod
    def from_dataset(cls, dataset: str, variant: str, *, expected_sha256: Optional[str] = None) -> 'Artifact':
        return cls.from_hf(
            hf_repo='macavaney/pyterrier-from-dataset',
            branch=f'{dataset}.{variant}',
            expected_sha256=expected_sha256)


from_url = Artifact.from_url
from_dataset = Artifact.from_dataset


def load_metadata(path: str) -> Dict:
    """
    Load the metadata file for the artifact at the specified path.

    Args:
        path: The path of the artifact on disk.

    Returns:
        The loaded metadata file, or an empty dictionary if the metadata file does not exist.

    Raises:
        FileNotFoundError: If the specified path does not exist.
    """
    if not os.path.isdir(path):
        raise FileNotFoundError(f'{path} not found')

    metadata_file = os.path.join(path, 'pt_meta.json')
    if os.path.exists(metadata_file):
        with open(metadata_file, 'rt') as f:
            return json.load(f)

    return {}


def _metadata_adapter(path):
    directory_listing = os.listdir(path)
    for entry_point in pta.io.entry_points('pyterrier.artifact.metadata_adapter'):
        if (metadata := entry_point.load()(path, directory_listing)) is not None:
            return metadata
    return {}


def load(path: str) -> Artifact:
    """Load the artifact from the specified path.

    This function attempts to load the artifact using a supporting Artifact implementation.

    Args:
        path: The path of the artifact on disk.

    Returns:
        The loaded artifact.

    Raises:
        FileNotFoundError: If the specified path does not exist.
        ValueError: If no implementation is found that supports the artifact at the specified path.
    """
    metadata = load_metadata(path)

    if 'type' not in metadata or 'format' not in metadata:
        metadata.update(_metadata_adapter(path))

    if 'type' in metadata and 'format' in metadata:
        entry_points = list(pta.io.entry_points('pyterrier.artifact'))
        matching_entry_points = [ep for ep in entry_points if ep.name == '{type}.{format}'.format(**metadata)]
        for entry_point in matching_entry_points:
            artifact_cls = entry_point.load()
            return artifact_cls(path)

    # coudn't find an implementation that supports this artifact
    error = [
        f'No implementation found that supports the artifact at {path_repr(path)}.',
        f'type={metadata.get("type", "(unknown)")}, format={metadata.get("format", "(unknown)")}.'
    ]
    if 'package_hint' in metadata:
        error.append(f'Do you need to `pip install {metadata["package_hint"]}`?')
    raise ValueError('\n'.join(error))


def build_package(artifact_path: str, package_path: Optional[str] = None, verbose: bool = True, max_file_size: Optional[float] = None) -> str:
    """
    Build a package from the specified artifact path.

    Packaged artifacts are useful for distributing an artifact as a single file, such as from Artifact.from_url().
    A separate metadata file is also generated, which gives information about the package's contents, including
    file sizes and an expected hash for the package.

    Args:
        artifact_path: The path of the artifact to package.
        package_path: The path of the package to create. Defaults to the artifact path with a .tar.lz4 extension.
        verbose: Whether to display a progress bar when packaging.

    Returns:
        The path of the package created.
    """
    if package_path is None:
        package_path = artifact_path + '.tar.lz4'

    metadata = {
        'expected_sha256': None,
        'total_size': 0,
        'contents': [],
    }

    chunk_num = 0
    chunk_start_offset = 0
    def manage_maxsize(_):
        nonlocal raw_fout
        nonlocal chunk_num
        nonlocal chunk_start_offset
        if max_file_size is not None and raw_fout.tell() >= max_file_size:
            raw_fout.flush()
            chunk_start_offset += raw_fout.tell()
            raw_fout.close()
            if chunk_num == 0:
                metadata['segments'] = []
                metadata['segments'].append({'idx': chunk_num, 'offset': 0})
            chunk_num += 1
            if verbose:
                print(f'starting segment {chunk_num}')
            metadata['segments'].append({'idx': chunk_num, 'offset': chunk_start_offset})
            raw_fout = stack.enter_context(pta.io.finalized_open(f'{package_path}.{chunk_num}', 'b'))
            sha256_fout.replace_writer(raw_fout)

    with contextlib.ExitStack() as stack:
        raw_fout = stack.enter_context(pta.io.finalized_open(f'{package_path}.{chunk_num}', 'b'))
        sha256_fout = stack.enter_context(pta.io.HashWriter(raw_fout))
        lz4_fout = stack.enter_context(LZ4FrameFile(sha256_fout, 'wb'))
        tarout = stack.enter_context(tarfile.open(fileobj=lz4_fout, mode='w'))

        for root, dirs, files in os.walk(artifact_path):
            rel_root = os.path.relpath(root, start=artifact_path)
            if rel_root != '.':
                tar_record = tarfile.TarInfo(rel_root)
                tar_record.type = tarfile.DIRTYPE
                tarout.addfile(tar_record)
            for file in sorted(files):
                lz4_fout.flush() # flush before each file, allowing seeking directly to this file within the archive
                file_full_path = os.path.join(root, file)
                file_rel_path = os.path.join(rel_root, file) if rel_root != '.' else file
                tar_record = tarfile.TarInfo(file_rel_path)
                tar_record.size = os.path.getsize(file_full_path)
                metadata['contents'].append({
                    'path': file_rel_path,
                    'size': tar_record.size,
                    'offset': chunk_start_offset + raw_fout.tell(),
                })
                metadata['total_size'] += tar_record.size
                if verbose:
                    print(f'adding {file_rel_path} [{pta.io.byte_count_to_human_readable(tar_record.size)}]')

                with open(file_full_path, 'rb') as fin, \
                     pta.io.CallbackReader(fin, manage_maxsize) as fin:
                    tarout.addfile(tar_record, fin)

        tarout.close()
        lz4_fout.close()

        metadata['expected_sha256'] = sha256_fout.hexdigest()

        metadata_out = stack.enter_context(pta.io.finalized_open(f'{package_path}.json', 't'))
        json.dump(metadata, metadata_out)
        metadata_out.write('\n')

    if chunk_num == 0:
        # no chunking was actually done, can use provided name directly
        os.rename(f'{package_path}.{chunk_num}', package_path)

    return package_path


def path_repr(path):
    artifact_info_path = f'{path}.json'
    if os.path.exists(artifact_info_path):
        with open(artifact_info_path) as fin:
            artifact_info = json.load(fin)
        if 'url' in artifact_info:
            return f'{path!r} <from {artifact_info["url"]!r}>'
        if 'path' in artifact_info:
            return f'{path!r} <from {artifact_info["path"]!r}>'
    return repr(path)



import os
import io
import tempfile
import shutil
import urllib
from abc import ABC, abstractmethod
from hashlib import sha256
from contextlib import contextmanager, ExitStack
from importlib.metadata import EntryPoint
from importlib.metadata import entry_points as eps
import pyterrier as pt
from typing import Optional, Tuple

DEFAULT_CHUNK_SIZE = 16_384 # 16kb


@contextmanager
def _finalized_open_base(path: str, mode: str, open_fn) -> io.IOBase:
    assert mode in ('b', 't') # must supply either binary or text mode
    prefix = f'.{os.path.basename(path)}.tmp.'
    dirname = os.path.dirname(path)
    try:
        fd, path_tmp = tempfile.mkstemp(prefix=prefix, dir=dirname)
        os.close(fd) # mkstemp returns a low-level file descriptor... Close it and re-open the file the normal way
        with open_fn(path_tmp, f'w{mode}') as fout:
            yield fout
        os.chmod(path_tmp, 0o666) # default file umask
    except:
        os.remove(path_tmp)
        raise

    os.replace(path_tmp, path)


def finalized_open(path: str, mode: str):
    """
    Opens a file for writing, but reverts it if there was an error in the process.

    Args:
        path(str): Path of file to open
        mode(str): Either t or b, for text or binary mode

    Example:
        Returns a contextmanager that provides a file object, so should be used in a "with" statement. E.g.::

            with pt.io.finalized_open("file.txt", "t") as f:
                f.write("some text")
            # file.txt exists with contents "some text"

        If there is an error when writing, the file is reverted::

            with pt.io.finalized_open("file.txt", "t") as f:
                f.write("some other text")
                raise Exception("an error")
            # file.txt remains unchanged (if existed, contents unchanged; if didn't exist, still doesn't)
    """
    return _finalized_open_base(path, mode, open)


@contextmanager
def finalized_directory(path: str) -> str:
    prefix = f'.{os.path.basename(path)}.tmp.'
    dirname = os.path.dirname(path)
    try:
        path_tmp = tempfile.mkdtemp(prefix=prefix, dir=dirname)
        yield path_tmp
        os.chmod(path_tmp, 0o777) # default directory umask
    except:
        try:
            shutil.rmtree(path_tmp)
        except:
            raise
        raise

    os.replace(path_tmp, path)


def download(url: str, path: str, *, expected_sha256: str = None, verbose: bool = True) -> None:
    with finalized_open(path) as fout, \
         download_stream(url, expected_sha256=expected_sha256, verbose=verbose) as fin:
        while chunk := fin.read1():
            fout.write(chunk)


@contextmanager
def download_stream(url: str, *, expected_sha256: Optional[str] = None, verbose: bool = True) -> io.IOBase:
    with ExitStack() as stack:
        fin = stack.enter_context(urllib.request.urlopen(url))
        if fin.status != 200:
            raise OSError(f'Unhandled status code: {fin.status}')

        if verbose:
            total = int(fin.headers.get('Content-Length', 0)) or None
            fin = stack.enter_context(TqdmReader(fin, total=total, desc=url))

        if expected_sha256 is not None:
            fin = stack.enter_context(HashReader(fin, expected=expected_sha256))

        yield fin


@contextmanager
def open_or_download_stream(
    path_or_url: str,
    *,
    expected_sha256: Optional[str] = None,
    verbose: bool = True
) -> io.IOBase:
    if path_or_url.startswith('http://') or path_or_url.startswith('https://'):
        with download_stream(path_or_url, expected_sha256=expected_sha256, verbose=verbose) as fin:
            yield fin
    elif os.path.isfile(path_or_url):
        with ExitStack() as stack:
            fin = stack.enter_context(open(path_or_url, 'rb'))

            if verbose:
                total = os.path.getsize(path_or_url)
                fin = stack.enter_context(TqdmReader(fin, total=total, desc=path_or_url))

            if expected_sha256 is not None:
                fin = stack.enter_context(HashReader(fin, expected=expected_sha256))

            yield fin
    else:
        raise OSError(f'path or url {path_or_url!r} not found')


class _NosyReader(io.BufferedIOBase, ABC):
    def __init__(self, reader: io.IOBase):
        self.reader = reader
        self.seek = self.reader.seek
        self.tell = self.reader.tell
        self.seekable = self.reader.seekable
        self.readable = self.reader.readable
        self.writable = self.reader.writable
        self.flush = self.reader.flush
        self.isatty = self.reader.isatty

    @abstractmethod
    def on_data(self, data: bytes) -> None:
        pass

    def read1(self, size: int = -1) -> bytes:
        if size == -1:
            size = DEFAULT_CHUNK_SIZE
        chunk = self.reader.read1(min(size, DEFAULT_CHUNK_SIZE))
        self.on_data(chunk)
        return chunk

    def read(self, size: int = -1) -> bytes:
        if size == -1:
            size = DEFAULT_CHUNK_SIZE
        chunk = self.reader.read(min(size, DEFAULT_CHUNK_SIZE))
        self.on_data(chunk)
        return chunk

    def close(self) -> None:
        self.reader.close()


class _NosyWriter(io.BufferedIOBase, ABC):
    def __init__(self, writer: io.IOBase):
        self.writer = writer
        self.seek = self.writer.seek
        self.tell = self.writer.tell
        self.seekable = self.writer.seekable
        self.readable = self.writer.readable
        self.writable = self.writer.writable
        self.flush = self.writer.flush
        self.isatty = self.writer.isatty
        self.close = self.writer.close
        self.sha256 = sha256()

    @abstractmethod
    def on_data(self, data: bytes) -> None:
        pass

    def write(self, data: bytes) -> None:
        self.writer.write(data)
        self.on_data(data)

    def replace_writer(self, writer: io.IOBase) -> None:
        self.writer = writer
        self.seek = self.writer.seek
        self.tell = self.writer.tell
        self.seekable = self.writer.seekable
        self.readable = self.writer.readable
        self.writable = self.writer.writable
        self.flush = self.writer.flush
        self.isatty = self.writer.isatty
        self.close = self.writer.close


class HashReader(_NosyReader):
    def __init__(self, reader: io.IOBase, *, hashfn=sha256, expected: Optional[str] = None):
        super().__init__(reader)
        self.hash = hashfn()
        self.expected = expected

    def on_data(self, data: bytes) -> None:
        self.hash.update(data)

    def hexdigest(self) -> str:
        return self.hash.hexdigest()

    def close(self) -> None:
        self.reader.close()
        if self.expected is not None:
            if self.expected.lower() != self.hexdigest():
                raise ValueError(f'Expected sha256 {self.expected!r} but found {self.hexdigest()!r}')


class HashWriter(_NosyWriter):
    def __init__(self, writer: io.IOBase, *, hashfn=sha256):
        super().__init__(writer)
        self.hash = hashfn()

    def on_data(self, data: bytes) -> None:
        self.hash.update(data)

    def hexdigest(self) -> str:
        return self.hash.hexdigest()


class TqdmReader(_NosyReader):
    def __init__(self, reader: io.IOBase, *, total: int = None, desc: str = None, disable: bool = False):
        super().__init__(reader)
        self.pbar = pt.tqdm(total=total, desc=desc, unit="B", unit_scale=True, unit_divisor=1024, disable=disable)

    def on_data(self, data: bytes) -> None:
        self.pbar.update(len(data))

    def close(self) -> None:
        super().close()
        self.reader.close()


class CallbackReader(_NosyReader):
    def __init__(self, reader: io.IOBase, callback):
        super().__init__(reader)
        self.callback = callback

    def on_data(self, data: bytes) -> None:
        self.callback(data)


class MultiReader(io.BufferedIOBase):
    def __init__(self, readers):
        self.readers = readers
        self._reader = next(self.readers)
        self.reader = self._reader.__enter__()
        self.seek = self.reader.seek
        self.tell = self.reader.tell
        self.seekable = self.reader.seekable
        self.readable = self.reader.readable
        self.writable = self.reader.writable
        self.flush = self.reader.flush
        self.isatty = self.reader.isatty
        self.close = self.reader.close

    def read1(self, size: int = -1) -> bytes:
        if size == -1:
            size = DEFAULT_CHUNK_SIZE
        chunk = self.reader.read1(min(size, DEFAULT_CHUNK_SIZE))
        if len(chunk) == 0:
            self.reader.close()
            try:
                self._reader = next(self.readers)
            except StopIteration:
                self._reader = None
                self.reader = None
                return chunk
            self.reader = self._reader.__enter__()
            self.pbar = self.reader.pbar
            self.seek = self.reader.seek
            self.tell = self.reader.tell
            self.seekable = self.reader.seekable
            self.readable = self.reader.readable
            self.writable = self.reader.writable
            self.flush = self.reader.flush
            self.isatty = self.reader.isatty
            self.close = self.reader.close
            chunk = self.reader.read1(min(size, DEFAULT_CHUNK_SIZE))
        return chunk

    def read(self, size: int = -1) -> bytes:
        chunk = b''
        if size == -1:
            size = DEFAULT_CHUNK_SIZE
        while len(chunk) < size and self.reader is not None:
            chunk += self.reader.read(size - len(chunk))
            if len(chunk) < size:
                self.reader.close()
                try:
                    self._reader = next(self.readers)
                except StopIteration:
                    self._reader = None
                    self.reader = None
                    return chunk
                self.reader = self._reader.__enter__()
                self.pbar = self.reader.pbar
                self.seek = self.reader.seek
                self.tell = self.reader.tell
                self.seekable = self.reader.seekable
                self.readable = self.reader.readable
                self.writable = self.reader.writable
                self.flush = self.reader.flush
                self.isatty = self.reader.isatty
                self.close = self.reader.close
        return chunk


def path_is_under_base(path: str, base: str) -> bool:
    return os.path.realpath(os.path.abspath(os.path.join(base, path))).startswith(os.path.realpath(base))


def byte_count_to_human_readable(byte_count: float) -> str:
    units = ['B', 'KB', 'MB', 'GB', 'TB']
    while byte_count > 1024 and len(units) > 1:
        byte_count /= 1024
        units = units[1:]
    if units[0] == 'B':
        return f'{byte_count:.0f} {units[0]}'
    return f'{byte_count:.1f} {units[0]}'


def entry_points(group: str) -> Tuple[EntryPoint, ...]:
    try:
        return tuple(eps(group=group))
    except TypeError:
        return tuple(eps().get(group, tuple()))


def pyterrier_home() -> str:
    if "PYTERRIER_HOME" in os.environ:
        home = os.environ["PYTERRIER_HOME"]
    else:
        home = os.path.expanduser('~/.pyterrier')
    if not os.path.exists(home):
        os.makedirs(home)
    return home



from types import TracebackType
from typing import List, Optional, Type

import pandas as pd


class TransformerMode:
    def __init__(self, missing_columns: List[str], extra_columns: List[str], mode_name: Optional[str] = None):
        self.missing_columns = missing_columns
        self.extra_columns = extra_columns
        self.mode_name = mode_name

    def __str__(self):
        return f'{self.mode_name} (missing: {self.missing_columns}, extra: {self.extra_columns})'

    def __repr__(self):
        return f'TransformerMode(missing_columns={self.missing_columns!r}, ' \
               f'extra_columns={self.extra_columns!r}, ' \
               f'mode_name={self.mode_name!r})'


class InputValidationError(KeyError):
    def __init__(self, message: str, modes: List[TransformerMode]):
        assert len(modes) > 0
        super().__init__(message)
        self.modes = modes

    def __str__(self):
        return self.args[0] + ' ' + str(self.modes)

    def __repr__(self):
        return f'InputValidationError({self.args[0]!r}, {self.modes!r})'


def columns(inp: pd.DataFrame,
            *,
            includes: Optional[List[str]] = None,
            excludes: Optional[List[str]] = None) -> None:
    """Check that the input frame has the expected columns.

    Raises:
        InputValidationError: If is missing any of the `includes` columns, or has any of the `excludes` columns.
    """
    with any(inp) as v:
        v.columns(includes=includes, excludes=excludes)


def query_frame(inp: pd.DataFrame, extra_columns: Optional[List[str]] = None) -> None:
    """Check that the input frame is a valid query frame.

    Raises:
        InputValidationError: If the input frame is not a valid query frame.
    """
    with any(inp) as v:
        v.query_frame(extra_columns)



def result_frame(inp: pd.DataFrame, extra_columns: Optional[List[str]] = None) -> None:
    """Check that the input frame is a valid result frame.

    Raises:
        InputValidationError: If the input frame is not a valid result frame.
    """
    with any(inp) as v:
        v.result_frame(extra_columns)


def document_frame(inp: pd.DataFrame, extra_columns: Optional[List[str]] = None) -> None:
    """Check that the input frame is a valid document frame.

    Raises:
        InputValidationError: If the input frame is not a valid document frame.
    """
    with any(inp) as v:
        v.document_frame(extra_columns)


def any(inp: pd.DataFrame) -> 'ValidationContextManager':
    return ValidationContextManager(inp)


class ValidationContextManager:
    def __init__(self, inp: pd.DataFrame):
        self.inp = inp
        self.mode = None
        self.attempts = 0
        self.errors = []

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> Optional[bool]:

        if exc_type is not None:
            return False # the captured exception takes priority

        if self.attempts > 0 and self.attempts == len(self.errors):
            raise InputValidationError("DataFrame does not match required columns for this transformer.", self.errors)

    def columns(self,
                *,
                includes: Optional[List[str]] = None,
                excludes: Optional[List[str]] = None,
                mode: str = None) -> bool:
        includes = includes if includes is not None else []
        excludes = excludes if excludes is not None else []
        missing_columns = set(includes) - set(self.inp.columns)
        extra_columns = set(excludes) & set(self.inp.columns)
        self.attempts += 1

        if missing_columns or extra_columns:
            self.errors.append(TransformerMode(
                missing_columns=[c for c in includes if c in missing_columns],
                extra_columns=[c for c in excludes if c in extra_columns],
                mode_name=mode,
            ))
            return False

        if self.mode is None and mode is not None:
            self.mode = mode

        return True

    def query_frame(self, extra_columns: Optional[List[str]] = None, mode: str = None) -> bool:
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['qid'] + extra_columns, excludes=['docno'], mode=mode)

    def result_frame(self, extra_columns: Optional[List[str]] = None, mode: str = None) -> bool:
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['qid', 'docno'] + extra_columns, mode=mode)

    def document_frame(self, extra_columns: Optional[List[str]] = None, mode: str = None) -> bool:
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['docno'] + extra_columns, excludes=['qid'], mode=mode)



from pyterrier_alpha import validate
from pyterrier_alpha.dataframe_builder import DataFrameBuilder
from pyterrier_alpha import io
from pyterrier_alpha import artifact
from pyterrier_alpha.artifact import Artifact
from pyterrier_alpha.rbo import RBO, rbo

__version__ = '0.3.0'



import setuptools

def get_version(path):
    for line in open(path, 'rt'):
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

setuptools.setup(
    name="pyterrier-alpha",
    version=get_version("pyterrier_alpha/__init__.py"),
    author="Sean MacAvaney",
    author_email='sean.macavaney@glasgow.ac.uk',
    description="Alpha channel of features for PyTerrier",
    long_description=open('README.md', 'rt').read(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
      'python-terrier',
    ],
    python_requires='>=3.6',
)
